import logging

import fontTools.designspaceLib as designspaceLib
import pytest
from fontTools.pens.recordingPen import RecordingPen

import ufo2ft.instantiator
from ufo2ft.util import openFont, openFontFactory


def test_interpolation_weight_width_class(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "MutatorSans" / "MutatorSans.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    for instance in designspace.instances:
        instance.font = generator.generate_instance(instance)

    # LightCondensed
    font = designspace.instances[0].font
    assert font.info.openTypeOS2WeightClass == 1
    assert font.info.openTypeOS2WidthClass == 1

    # BoldCondensed
    font = designspace.instances[1].font
    assert font.info.openTypeOS2WeightClass == 1000
    assert font.info.openTypeOS2WidthClass == 1

    # LightWide
    font = designspace.instances[2].font
    assert font.info.openTypeOS2WeightClass == 1
    assert font.info.openTypeOS2WidthClass == 9

    # BoldWide
    font = designspace.instances[3].font
    assert font.info.openTypeOS2WeightClass == 1000
    assert font.info.openTypeOS2WidthClass == 9

    # Medium_Narrow_I
    font = designspace.instances[4].font
    assert font.info.openTypeOS2WeightClass == 500
    assert font.info.openTypeOS2WidthClass == 9

    # Medium_Wide_I
    font = designspace.instances[5].font
    assert font.info.openTypeOS2WeightClass == 500
    assert font.info.openTypeOS2WidthClass == 9

    # Two
    font = designspace.instances[6].font
    assert font.info.openTypeOS2WeightClass == 1000
    assert font.info.openTypeOS2WidthClass == 9

    # One
    font = designspace.instances[7].font
    assert font.info.openTypeOS2WeightClass == 500
    assert font.info.openTypeOS2WidthClass == 9


def test_default_groups_only(ufo_module, data_dir, caplog):
    """Test that only the default source's groups end up in instances."""

    d = designspaceLib.DesignSpaceDocument()
    d.addAxisDescriptor(
        name="Weight", tag="wght", minimum=300, default=300, maximum=900
    )
    d.addSourceDescriptor(location={"Weight": 300}, font=ufo_module.Font())
    d.addSourceDescriptor(location={"Weight": 900}, font=ufo_module.Font())
    d.addInstanceDescriptor(styleName="2", location={"Weight": 400})

    d.sources[0].font.groups["public.kern1.GRK_alpha_alt_LC_1ST"] = [
        "alpha.alt",
        "alphatonos.alt",
    ]
    d.sources[1].font.groups["public.kern1.GRK_alpha_LC_1ST"] = [
        "alpha.alt",
        "alphatonos.alt",
    ]

    generator = ufo2ft.instantiator.Instantiator.from_designspace(d)
    assert "contains different groups than the default source" in caplog.text

    instance = generator.generate_instance(d.instances[0])
    assert instance.groups == {
        "public.kern1.GRK_alpha_alt_LC_1ST": ["alpha.alt", "alphatonos.alt"]
    }


def test_default_groups_only2(ufo_module, data_dir, caplog):
    """Test that the group difference warning is not triggered if non-default
    source groups are empty."""

    d = designspaceLib.DesignSpaceDocument()
    d.addAxisDescriptor(
        name="Weight", tag="wght", minimum=300, default=300, maximum=900
    )
    d.addSourceDescriptor(location={"Weight": 300}, font=ufo_module.Font())
    d.addSourceDescriptor(location={"Weight": 900}, font=ufo_module.Font())
    d.addInstanceDescriptor(styleName="2", location={"Weight": 400})

    d.sources[0].font.groups["public.kern1.GRK_alpha_alt_LC_1ST"] = [
        "alpha.alt",
        "alphatonos.alt",
    ]

    generator = ufo2ft.instantiator.Instantiator.from_designspace(d)
    assert "contains different groups than the default source" not in caplog.text

    instance = generator.generate_instance(d.instances[0])
    assert instance.groups == {
        "public.kern1.GRK_alpha_alt_LC_1ST": ["alpha.alt", "alphatonos.alt"]
    }


def test_interpolation_no_rounding(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "MutatorSans" / "MutatorSans.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    designspace.instances[4].location = {"weight": 123.456, "width": 789.123}
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=False
    )

    instance_font = generator.generate_instance(designspace.instances[4])
    assert isinstance(instance_font.info.ascender, float)
    assert isinstance(instance_font.kerning[("A", "J")], float)
    assert isinstance(instance_font["A"].contours[0][0].x, float)


def test_interpolation_rounding(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "MutatorSans" / "MutatorSans.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    designspace.instances[4].location = {"weight": 123.456, "width": 789.123}
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace.instances[4])
    assert isinstance(instance_font.info.ascender, int)
    assert isinstance(instance_font.kerning[("A", "J")], int)
    assert isinstance(instance_font["A"].contours[0][0].x, int)


def test_weight_class_from_wght_axis():
    assert ufo2ft.instantiator.weight_class_from_wght_value(-500) == 1
    assert ufo2ft.instantiator.weight_class_from_wght_value(1.1) == 1
    assert ufo2ft.instantiator.weight_class_from_wght_value(1) == 1
    assert ufo2ft.instantiator.weight_class_from_wght_value(500.6) == 501
    assert ufo2ft.instantiator.weight_class_from_wght_value(1000) == 1000
    assert ufo2ft.instantiator.weight_class_from_wght_value(1000.0) == 1000
    assert ufo2ft.instantiator.weight_class_from_wght_value(1000.1) == 1000
    assert ufo2ft.instantiator.weight_class_from_wght_value(2000.1) == 1000


def test_width_class_from_wdth_axis():
    assert ufo2ft.instantiator.width_class_from_wdth_value(-500) == 1
    assert ufo2ft.instantiator.width_class_from_wdth_value(50) == 1
    assert ufo2ft.instantiator.width_class_from_wdth_value(62.5) == 2
    assert ufo2ft.instantiator.width_class_from_wdth_value(75) == 3
    assert ufo2ft.instantiator.width_class_from_wdth_value(87.5) == 4
    assert ufo2ft.instantiator.width_class_from_wdth_value(100) == 5
    assert ufo2ft.instantiator.width_class_from_wdth_value(112) == 6
    assert ufo2ft.instantiator.width_class_from_wdth_value(112.5) == 6
    assert ufo2ft.instantiator.width_class_from_wdth_value(125) == 7
    assert ufo2ft.instantiator.width_class_from_wdth_value(130) == 7
    assert ufo2ft.instantiator.width_class_from_wdth_value(150) == 8
    assert ufo2ft.instantiator.width_class_from_wdth_value(190) == 9
    assert ufo2ft.instantiator.width_class_from_wdth_value(200) == 9
    assert ufo2ft.instantiator.width_class_from_wdth_value(1000) == 9


def test_swap_glyph_names(ufo_module, data_dir):
    ufo = openFont(data_dir / "SwapGlyphNames" / "A.ufo", ufo_module=ufo_module)

    ufo2ft.instantiator.swap_glyph_names(ufo, "a", "a.swap")

    # Test swapped outlines.
    assert ufo["a"].unicode == 0x61
    assert len(ufo["a"]) == 1
    assert len(ufo["a"][0]) == 8
    assert ufo["a"].width == 666
    assert ufo["a.swap"].unicode is None
    assert len(ufo["a.swap"]) == 1
    assert len(ufo["a.swap"][0]) == 4
    assert ufo["a.swap"].width == 600

    # Test swapped components.
    assert sorted(c.baseGlyph for c in ufo["aaa"].components) == [
        "a.swap",
        "a.swap",
        "x",
    ]
    assert sorted(c.baseGlyph for c in ufo["aaa.swap"].components) == ["a", "a", "y"]

    # Test swapped anchors.
    assert [dict(a) for a in ufo["a"].anchors] == [
        dict(x=153, y=0, name="bottom"),
        dict(x=153, y=316, name="top"),
    ]
    assert [dict(a) for a in ufo["a.swap"].anchors] == [
        dict(x=351, y=0, name="bottom"),
        dict(x=351, y=613, name="top"),
    ]

    # Test swapped glyph kerning.
    assert ufo.kerning == {
        ("public.kern1.a", "x"): 10,
        ("public.kern1.aswap", "x"): 20,
        ("a", "y"): 40,
        ("a.swap", "y"): 30,
        ("y", "a"): 60,
        ("y", "a.swap"): 50,
    }

    # Test swapped group membership.
    assert ufo.groups == {
        "public.kern1.a": ["a.swap"],
        "public.kern1.aswap": ["a"],
        "public.kern2.a": ["a.swap", "a"],
    }

    # Swap a second time.
    ufo2ft.instantiator.swap_glyph_names(ufo, "aaa", "aaa.swap")

    # Test swapped glyphs.
    assert sorted(c.baseGlyph for c in ufo["aaa"].components) == ["a", "a", "y"]
    assert sorted(c.baseGlyph for c in ufo["aaa.swap"].components) == [
        "a.swap",
        "a.swap",
        "x",
    ]

    # Test for no leftover temporary glyphs.
    assert {g.name for g in ufo} == {
        "space",
        "a",
        "a.swap",
        "aaa",
        "aaa.swap",
        "x",
        "y",
    }

    with pytest.raises(ufo2ft.instantiator.InstantiatorError, match="Cannot swap"):
        ufo2ft.instantiator.swap_glyph_names(ufo, "aaa", "aaa.swapa")


def test_swap_glyph_names_spec(ufo_module, data_dir):
    """Test that the rule example in the designspaceLib spec works.

    `adieresis` should look the same as before the rule application.

    [1]: fonttools/Doc/source/designspaceLib#ufo-instances
    """
    ufo = openFont(data_dir / "SwapGlyphNames" / "B.ufo", ufo_module=ufo_module)
    ufo2ft.instantiator.swap_glyph_names(ufo, "a", "a.alt")

    assert sorted(c.baseGlyph for c in ufo["adieresis"].components) == [
        "a.alt",
        "dieresiscomb",
    ]
    assert sorted(c.baseGlyph for c in ufo["adieresis.alt"].components) == [
        "a",
        "dieresiscomb",
    ]


def test_rules_are_applied_deterministically(ufo_module, data_dir):
    """Test that a combination of designspace rules that end up mapping
    serveral input glyphs to the same destination glyph result in a correct and
    deterministic series of glyph swaps.

    The example is a font with 2 Q designs that depend on a style axis
        style < 0.5: Q        style >= 0.5: Q.ss01
    and each Q also has an alternative shape in bolder weights (like Skia)
        weight < 780: Q       weight >= 780: Q.alt
        weight < 730: Q.ss01  weight >= 730: Q.ss01.alt

    Then we generate an instance at style = 1, weight = 900. From the rules,
    the default CMAP entry for Q should have the outlines of Q.ss01.alt from
    the black UFO.
    """
    doc = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceRuleOrder" / "MyFont.designspace"
    )
    doc.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(doc)
    instance = generator.generate_instance(doc.instances[0])
    pen = RecordingPen()
    instance["Q"].draw(pen)
    instance_recording = pen.value

    black_ufo = openFont(
        data_dir / "DesignspaceRuleOrder" / "MyFont_Black.ufo", ufo_module=ufo_module
    )
    pen = RecordingPen()
    black_ufo["Q.ss01.alt"].draw(pen)
    black_ufo_recording = pen.value

    assert instance_recording == black_ufo_recording


def test_raise_no_default_master(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "MutatorSans" / "MutatorSans_no_default.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))

    with pytest.raises(ufo2ft.instantiator.InstantiatorError, match="no default"):
        ufo2ft.instantiator.Instantiator.from_designspace(
            designspace, round_geometry=True
        )


def test_raise_failed_glyph_interpolation(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceBrokenTest" / "DesignspaceTest.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(designspace)

    with pytest.raises(
        ufo2ft.instantiator.InstantiatorError, match="Failed to generate instance"
    ):
        for instance in designspace.instances:
            instance.font = generator.generate_instance(instance)


def test_ignore_failed_glyph_interpolation(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceBrokenTest" / "DesignspaceTest.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(designspace)
    generator.skip_export_glyphs.append("asas")

    for instance in designspace.instances:
        instance.font = generator.generate_instance(instance)
        assert (
            not instance.font["asas"].contours and not instance.font["asas"].components
        )


def test_raise_anisotropic_location(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "MutatorSans" / "MutatorSans-width-only.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    designspace.instances[0].location["width"] = (100, 900)

    with pytest.raises(
        ufo2ft.instantiator.InstantiatorError, match="anisotropic instance locations"
    ):
        ufo2ft.instantiator.Instantiator.from_designspace(
            designspace, round_geometry=True
        )


def test_copy_nonkerning_group(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(designspace)

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font.groups == {
        "nonkerning_group": ["A"],
        "public.kern2.asdf": ["A"],
    }


def test_interpolation(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font["l"].width == 220


def test_interpolation_only_default(ufo_module, data_dir, caplog):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "MutatorSans" / "MutatorSans.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    for name in designspace.default.font.glyphOrder:
        if name != "A":
            del designspace.default.font[name]

    with caplog.at_level(logging.WARNING):
        generator = ufo2ft.instantiator.Instantiator.from_designspace(
            designspace, round_geometry=True
        )
    assert "contains glyphs that are missing from the" in caplog.text

    instance_font = generator.generate_instance(designspace.instances[0])
    assert {g.name for g in instance_font} == {"A"}


def test_interpolation_masters_as_instances(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir
        / "DesignspaceBrokenTest"
        / "Designspace-MastersAsInstances.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font.info.styleName == "Light ASDF"
    assert instance_font["l"].width == 160
    instance_font = generator.generate_instance(designspace.instances[1])
    assert instance_font.info.styleName == "Bold ASDF"
    assert instance_font["l"].width == 280


def test_non_default_layer(ufo_module, data_dir, caplog):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "MutatorSans" / "MutatorSans-non-default-layer.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )
    instance_font = generator.generate_instance(designspace.instances[0])
    assert {g.name for g in instance_font} == {"A", "S", "W"}


def test_instance_attributes(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest-instance-attrs.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font.info.familyName == "aaa"
    assert instance_font.info.styleName == "sss"
    assert instance_font.info.postscriptFontName == "ppp"
    assert instance_font.info.styleMapFamilyName == "yyy"
    assert instance_font.info.styleMapStyleName == "xxx"


def test_instance_no_attributes(ufo_module, data_dir, caplog):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest-bare.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    with caplog.at_level(logging.WARNING):
        instance_font = generator.generate_instance(designspace.instances[0])
    assert "missing the stylename attribute" in caplog.text

    assert instance_font.info.familyName == "MyFont"
    assert instance_font.info.styleName == "Light"
    assert instance_font.info.postscriptFontName is None
    assert instance_font.info.styleMapFamilyName is None
    assert instance_font.info.styleMapStyleName is None


def test_instance_lib_attributes(ufo_module, data_dir, caplog):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "MutatorSans" / "MutatorSans.designspace"
    )
    designspace.instances[0].lib["public.fontInfo"] = {
        "openTypeOS2Panose": [
            2,
            11,
            5,
            4,
            2,
            2,
            2,
            2,
            2,
            4,
        ],
        "invalidFontInfoAttribute": "foobarbaz",
    }
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    with caplog.at_level(logging.WARNING):
        instance_font = generator.generate_instance(designspace.instances[0])

    assert instance_font.info.openTypeOS2Panose == [2, 11, 5, 4, 2, 2, 2, 2, 2, 4]
    assert (
        "Font 'MutatorMathTest' at instance location {'width': 0.0, 'weight': 0.0} "
        "has an unknown font info attribute 'invalidFontInfoAttribute' "
        "with value foobarbaz. This will be ignored."
    ) in caplog.text

    instance_font2 = generator.generate_instance(designspace.instances[1])
    assert instance_font2.info.openTypeOS2Panose is None


def test_axis_mapping(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest-wght-wdth.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font.info.openTypeOS2WeightClass == 400
    assert instance_font.info.openTypeOS2WidthClass == 5
    assert instance_font.info.italicAngle is None
    assert instance_font.lib["designspace.location"] == [
        ("weight", 100.0),
        ("width", 100.0),
    ]


def test_axis_mapping_manual_os2_classes(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest-wght-wdth.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    designspace.sources[0].font.info.openTypeOS2WeightClass = 800
    designspace.sources[0].font.info.openTypeOS2WidthClass = 7
    designspace.sources[1].font.info.openTypeOS2WeightClass = 900
    designspace.sources[1].font.info.openTypeOS2WidthClass = 9
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font.info.openTypeOS2WeightClass == 850
    assert instance_font.info.openTypeOS2WidthClass == 8
    assert instance_font.info.italicAngle is None
    assert instance_font.lib["designspace.location"] == [
        ("weight", 100.0),
        ("width", 100.0),
    ]


def test_axis_mapping_no_os2_width_class_inference(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest-bare.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    designspace.sources[0].font.info.openTypeOS2WeightClass = 800
    designspace.sources[1].font.info.openTypeOS2WeightClass = 900
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font.info.openTypeOS2WeightClass == 850
    assert instance_font.info.openTypeOS2WidthClass is None
    assert instance_font.info.italicAngle is None
    assert instance_font.lib["designspace.location"] == [("weight", 100.0)]


def test_axis_mapping_no_os2_class_inference(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest-opsz.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font.info.openTypeOS2WeightClass is None
    assert instance_font.info.openTypeOS2WidthClass is None
    assert instance_font.info.italicAngle is None
    assert instance_font.lib["designspace.location"] == [("optical", 15.0)]


def test_axis_mapping_italicAngle_inference(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest-slnt.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font.info.openTypeOS2WeightClass is None
    assert instance_font.info.openTypeOS2WidthClass is None
    assert instance_font.info.italicAngle == 40.123
    assert instance_font.lib["designspace.location"] == [("slant", 40.123)]


def test_lib_into_instance(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest-lib.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    assert designspace.default.font.lib["blorb"] == "asasa"
    assert "public.skipExportGlyphs" not in designspace.sources[0].font.lib

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font.lib["blorb"] == "asasa"
    assert instance_font.lib["public.skipExportGlyphs"] == ["a", "b", "c"]

    instance_font2 = generator.generate_instance(designspace.instances[1])
    assert instance_font2.lib["blorb"] == "asasa"
    assert instance_font2.lib["public.skipExportGlyphs"] == ["a", "b", "c"]


def test_font_info_fallback_in_instance(ufo_module, data_dir):
    """Test that font info attributes are correctly copied according to the designspace
    docs fallback rules [1]. Notes: language "de" is 0x0407/1031.

    [1]:https://fonttools.readthedocs.io/en/stable/designspaceLib/index.html#common-lib-key-registry
    """
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest-name.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    def assert_name_record_in_list(name_record, name_record_list):
        matched_name_records = [
            x
            for x in name_record_list
            if x["platformID"] == name_record["platformID"]
            and x["encodingID"] == name_record["encodingID"]
            and x["languageID"] == name_record["languageID"]
            and x["nameID"] == name_record["nameID"]
        ]
        # only one name record should match with given platformID,
        # encodingID, languageID, nameID
        assert len(matched_name_records) == 1
        # and the string value should match
        assert matched_name_records[0]["string"] == name_record["string"]

    # font info from the default source
    assert designspace.default.font.info.familyName == "MyFont"
    assert designspace.default.font.info.styleName == "Light"
    assert designspace.default.font.info.xHeight == 0
    assert designspace.default.font.info.openTypeNameRecords is None
    assert designspace.default.font.info.openTypeNameSampleText is None
    assert designspace.default.font.info.postscriptFontName == "MyFont-Light"
    # font info from the designspace root lib
    assert designspace.lib["public.fontInfo"]["familyName"] == "MyFont VF"
    assert designspace.lib["public.fontInfo"]["styleName"] == "Regular"
    assert designspace.lib["public.fontInfo"]["versionMajor"] == 1
    assert designspace.lib["public.fontInfo"]["versionMinor"] == 0
    assert designspace.lib["public.fontInfo"]["xHeight"] == 50
    assert "openTypeNameSampleText" not in designspace.lib["public.fontInfo"]
    assert_name_record_in_list(
        {
            "platformID": 3,
            "encodingID": 1,
            "languageID": 1031,
            "nameID": 0,
            "string": "Urheberrecht 2025",
        },
        designspace.lib["public.fontInfo"]["openTypeNameRecords"],
    )

    # instance 0 has no instance font info, should use root lib attributes
    assert designspace.instances[0].lib == {}
    assert designspace.instances[0].familyName is None
    assert designspace.instances[0].styleName is None
    instance_light_font = generator.generate_instance(designspace.instances[0])
    assert instance_light_font.info.familyName == "MyFont VF"
    assert instance_light_font.info.styleName == "Regular"
    assert instance_light_font.info.versionMajor == 1
    assert instance_light_font.info.versionMinor == 0
    assert instance_light_font.info.xHeight == 50
    assert instance_light_font.info.openTypeNameSampleText is None
    # although the default source defines a 'postscriptFontName', this is considered
    # instance-specific and does not be copied (unless when there's only one source
    # and can unambiguously be copied to all instances, as they'll all be at the
    # default location, see `test_static_font_default_instance_inherits_all_fontinfo`
    assert instance_light_font.info.postscriptFontName is None
    assert_name_record_in_list(
        {
            "platformID": 3,
            "encodingID": 1,
            "languageID": 1031,
            "nameID": 0,
            "string": "Urheberrecht 2025",
        },
        instance_light_font.info.openTypeNameRecords,
    )

    # instance 4 add new fontinfo attributes
    assert (
        designspace.instances[4].lib["public.fontInfo"]["openTypeNameSampleText"]
        == "Sample text for Bold instance"
    )
    assert designspace.instances[4].familyName == "MyFont"
    assert designspace.instances[4].styleName == "Bold"
    assert designspace.instances[4].localisedFamilyName == {"de": "MehrNamenTest"}
    assert designspace.instances[4].localisedStyleName == {}
    assert designspace.instances[4].localisedStyleMapFamilyName == {}
    assert designspace.instances[4].localisedStyleMapStyleName == {}
    assert {
        "platformID": 3,
        "encodingID": 1,
        "languageID": 1031,
        "nameID": 19,
        "string": "Beispieltext für die Fettschrift",
    } in designspace.instances[4].lib["public.fontInfo"]["openTypeNameRecords"]
    instance_bold_font = generator.generate_instance(designspace.instances[4])
    assert instance_bold_font.info.familyName == "MyFont"
    assert instance_bold_font.info.styleName == "Bold"
    assert (
        instance_bold_font.info.openTypeNameSampleText
        == "Sample text for Bold instance"
    )
    # check if lib-public.fontInfo-openTypeNameRecords is merged correctly
    assert_name_record_in_list(
        {
            "platformID": 3,
            "encodingID": 1,
            "languageID": 1031,
            "nameID": 0,
            "string": "Urheberrecht 2025",
        },
        instance_bold_font.info.openTypeNameRecords,
    )
    assert_name_record_in_list(
        {
            "platformID": 3,
            "encodingID": 1,
            "languageID": 1031,
            "nameID": 1,
            "string": "MehrNamenTest",
        },
        instance_bold_font.info.openTypeNameRecords,
    )
    assert_name_record_in_list(
        {
            "platformID": 3,
            "encodingID": 1,
            "languageID": 1031,
            "nameID": 16,
            "string": "MehrNamenTest",
        },
        instance_bold_font.info.openTypeNameRecords,
    )
    assert_name_record_in_list(
        {
            "platformID": 3,
            "encodingID": 1,
            "languageID": 1031,
            "nameID": 19,
            "string": "Beispieltext für die Fettschrift",
        },
        instance_bold_font.info.openTypeNameRecords,
    )

    # instance 1 has font info and multilingual familyName attributes,
    # but stylemapfamilyname not in, should autogenerate from familyName
    assert designspace.instances[1].familyName == "MyFont"
    assert designspace.instances[1].styleName == "SemiLight"
    assert designspace.instances[1].localisedFamilyName == {"de": "MehrNamenTest"}
    assert designspace.instances[1].localisedStyleName == {"de": "HalbLeicht"}
    assert designspace.instances[1].localisedStyleMapFamilyName == {}
    assert designspace.instances[1].localisedStyleMapStyleName == {}
    instance_semilight_font = generator.generate_instance(designspace.instances[1])
    assert instance_semilight_font.info.familyName == "MyFont"
    assert instance_semilight_font.info.styleName == "SemiLight"
    assert_name_record_in_list(
        {
            "platformID": 3,
            "encodingID": 1,
            "languageID": 0x0407,
            "nameID": 16,
            "string": "MehrNamenTest",
        },
        instance_semilight_font.info.openTypeNameRecords,
    )
    assert_name_record_in_list(
        {
            "platformID": 3,
            "encodingID": 1,
            "languageID": 0x0407,
            "nameID": 17,
            "string": "HalbLeicht",
        },
        instance_semilight_font.info.openTypeNameRecords,
    )
    assert_name_record_in_list(
        {
            "platformID": 3,
            "encodingID": 1,
            "languageID": 0x0407,
            "nameID": 1,
            "string": "MehrNamenTest HalbLeicht",
        },
        instance_semilight_font.info.openTypeNameRecords,
    )
    # check if lib-public.fontInfo-openTypeNameRecords is merged correctly
    assert_name_record_in_list(
        {
            "platformID": 3,
            "encodingID": 1,
            "languageID": 1031,
            "nameID": 0,
            "string": "Urheberrecht 2025",
        },
        instance_semilight_font.info.openTypeNameRecords,
    )

    # instance 2 has font info and all multilingual familyName attributes
    assert designspace.instances[2].localisedFamilyName == {"de": "MehrNamenTest"}
    assert designspace.instances[2].localisedStyleName == {"de": "DemiLeicht"}
    assert designspace.instances[2].localisedStyleMapFamilyName == {
        "de": "MehrNamenTest DemiLeichtx"
    }
    assert designspace.instances[2].localisedStyleMapStyleName == {"de": "Regular"}
    instance_demilight_font = generator.generate_instance(designspace.instances[2])
    assert len(instance_demilight_font.info.openTypeNameRecords) == 5
    assert_name_record_in_list(
        {
            "platformID": 3,
            "encodingID": 1,
            "languageID": 0x0407,
            "nameID": 16,
            "string": "MehrNamenTest",
        },
        instance_demilight_font.info.openTypeNameRecords,
    )
    assert_name_record_in_list(
        {
            "platformID": 3,
            "encodingID": 1,
            "languageID": 0x0407,
            "nameID": 17,
            "string": "DemiLeicht",
        },
        instance_demilight_font.info.openTypeNameRecords,
    )
    assert_name_record_in_list(
        {
            "platformID": 3,
            "encodingID": 1,
            "languageID": 0x0407,
            "nameID": 1,
            "string": "MehrNamenTest DemiLeichtx",
        },
        instance_demilight_font.info.openTypeNameRecords,
    )
    assert_name_record_in_list(
        {
            "platformID": 3,
            "encodingID": 1,
            "languageID": 0x0407,
            "nameID": 2,
            "string": "Regular",
        },
        instance_demilight_font.info.openTypeNameRecords,
    )
    # check if lib-public.fontInfo-openTypeNameRecords is merged correctly
    assert_name_record_in_list(
        {
            "platformID": 3,
            "encodingID": 1,
            "languageID": 1031,
            "nameID": 0,
            "string": "Urheberrecht 2025",
        },
        instance_demilight_font.info.openTypeNameRecords,
    )

    # instance 3 introduce fontinfo overrides
    assert designspace.instances[3].familyName == "MyFont"
    assert designspace.instances[3].styleName == "Regular"
    assert designspace.instances[3].lib["public.fontInfo"]["versionMajor"] == 3
    assert designspace.instances[3].lib["public.fontInfo"]["versionMinor"] == 141
    assert designspace.instances[3].lib["public.fontInfo"]["xHeight"] == 70
    assert designspace.instances[3].localisedFamilyName == {"de": "MehrNamenTest"}
    assert designspace.instances[3].localisedStyleName == {}
    assert designspace.instances[3].localisedStyleMapFamilyName == {}
    assert designspace.instances[3].localisedStyleMapStyleName == {}
    assert {
        "platformID": 3,
        "encodingID": 1,
        "languageID": 1031,
        "nameID": 0,
        "string": "Urheberrecht 2025 Overriden",
    } in designspace.instances[3].lib["public.fontInfo"]["openTypeNameRecords"]
    instance_regular_font = generator.generate_instance(designspace.instances[3])
    assert instance_regular_font.info.familyName == "MyFont"
    assert instance_regular_font.info.styleName == "Regular"
    assert instance_regular_font.info.versionMajor == 3
    assert instance_regular_font.info.versionMinor == 141
    assert instance_regular_font.info.xHeight == 70
    # check if lib-public.fontInfo-openTypeNameRecords is merged correctly
    # same platformID, encodingID, languageID, nameID should only exist once
    assert len(instance_regular_font.info.openTypeNameRecords) == 3
    assert_name_record_in_list(
        {
            "platformID": 3,
            "encodingID": 1,
            "languageID": 1031,
            "nameID": 0,
            "string": "Urheberrecht 2025 Overriden",
        },
        instance_regular_font.info.openTypeNameRecords,
    )
    assert_name_record_in_list(
        {
            "platformID": 3,
            "encodingID": 1,
            "languageID": 1031,
            "nameID": 1,
            "string": "MehrNamenTest",
        },
        instance_regular_font.info.openTypeNameRecords,
    )
    assert_name_record_in_list(
        {
            "platformID": 3,
            "encodingID": 1,
            "languageID": 1031,
            "nameID": 16,
            "string": "MehrNamenTest",
        },
        instance_regular_font.info.openTypeNameRecords,
    )


def test_font_info_fallback_should_skip_english_in_localized(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "BagelFatOne-Regular.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))

    # fontmake calls `splitInterpolable` before constructing the Instantiator.
    # It does that to split a v5 designspace which may contain discrete axes into
    # 'interpolable' sub-designspaces. This code also handles generating instance
    # names from STAT-related elements like axis <labelname>. As part of this, it
    # can sometimes add "en" (English) names to the sub-doc instances' localized
    # names (English names are normally defined in the instance attributes, but
    # subsequent designspaceLib code wants to access all the instance names via
    # the same localized* interface).
    # The uf2ft code in Instantiator._generate_instance_info that handles
    # the localized instance names currently does not distinguish between the
    # true localized names and these "en" name duplicates, and ends up
    # writing unnecessary openTypeNameRecords for the latter.
    [(_, designspace2)] = designspaceLib.split.splitInterpolable(designspace)
    assert designspace.instances[0].localisedFamilyName == {}
    assert designspace2.instances[0].localisedFamilyName == {"en": "Bagel Fat One"}

    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace2, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace2.instances[0])

    assert instance_font.info.familyName == "Bagel Fat One"
    assert instance_font.info.styleName == "Regular"
    assert instance_font.info.openTypeNameRecords is None


def test_data_independence(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )
    instance_font1 = generator.generate_instance(designspace.instances[0])
    designspace.instances[0].lib["aaaaaaaa"] = 1
    instance_font2 = generator.generate_instance(designspace.instances[0])

    instance_font1["l"].unicodes.append(2)
    assert instance_font1["l"].unicodes == [0x6C, 2]
    assert instance_font2["l"].unicodes == [0x6C]

    instance_font1["l"].lib["asdf"] = 1
    assert instance_font1["l"].lib == {"asdf": 1}
    assert not instance_font2["l"].lib

    generator.copy_lib["sdjkhsjdhjdf"] = 1
    instance_font1.lib["asdf"] = 1
    assert instance_font1.lib == {
        "asdf": 1,
        "blorb": "asasa",
        "designspace.location": [("weight", 100.0)],
        "public.skipExportGlyphs": [],
    }
    assert instance_font2.lib == {
        "blorb": "asasa",
        "designspace.location": [("weight", 100.0)],
        "public.skipExportGlyphs": [],
    }

    assert generator.copy_info.openTypeOS2Panose == [2, 11, 5, 4, 2, 2, 2, 2, 2, 4]
    generator.copy_info.openTypeOS2Panose.append(1000)
    assert instance_font1.info.openTypeOS2Panose == [2, 11, 5, 4, 2, 2, 2, 2, 2, 4]
    assert instance_font2.info.openTypeOS2Panose == [2, 11, 5, 4, 2, 2, 2, 2, 2, 4]

    # copy_feature_text not tested because it is a(n immutable) string

    assert not generator.skip_export_glyphs
    generator.skip_export_glyphs.extend(["a", "b"])
    assert not instance_font1.lib["public.skipExportGlyphs"]
    assert not instance_font2.lib["public.skipExportGlyphs"]
    instance_font1.lib["public.skipExportGlyphs"].append("z")
    assert not instance_font2.lib["public.skipExportGlyphs"]


def test_skipped_fontinfo_attributes():
    """Test that we consider all available font info attributes for copying."""
    import fontMath.mathInfo
    import fontTools.ufoLib

    SKIPPED_ATTRS = {
        "guidelines",
        "macintoshFONDFamilyID",
        "macintoshFONDName",
        "openTypeNameCompatibleFullName",
        "openTypeNamePreferredFamilyName",
        "openTypeNamePreferredSubfamilyName",
        "openTypeNameUniqueID",
        "openTypeNameWWSFamilyName",
        "openTypeNameWWSSubfamilyName",
        "postscriptFontName",
        "postscriptFullName",
        "postscriptUniqueID",
        "styleMapFamilyName",
        "styleMapStyleName",
        "styleName",
        "woffMetadataUniqueID",
        "year",
    }

    assert (
        fontTools.ufoLib.fontInfoAttributesVersion3
        - set(fontMath.mathInfo._infoAttrs.keys())
        - {"postscriptWeightName"}  # Handled in fontMath specially.
        - ufo2ft.instantiator.UFO_INFO_ATTRIBUTES_TO_COPY_TO_INSTANCES
        == SKIPPED_ATTRS
    )


def test_static_font_default_instance_inherits_all_fontinfo(data_dir, ufo_module):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest-static.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))

    assert len(designspace.sources) == 1  # single-master, static font
    assert len(designspace.instances) == 1
    # the default source's fontinfo.plist has a `postscriptFontName`
    designspace.sources[0].font.info.postscriptFontName = "MyFont-Light"

    # the DS instance does not have a `postScriptFontName` attribute, nor
    # does it have a `public.fontInfo` dict
    assert designspace.instances[0].postScriptFontName is None
    assert "public.fontInfo" not in designspace.instances[0].lib

    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )
    instance_font = generator.generate_instance(designspace.instances[0])

    # the instance inherits the postscriptFontName from the default source's
    # fontinfo.plist
    assert instance_font.info.postscriptFontName == "MyFont-Light"


def test_sparse_master_instance_at_non_default_location(ufo_module):
    """Designspaces with sparse/virtual masters (e.g. those Glyphs.app emits for a
    "Virtual Master" custom parameter) have only one source carrying fontinfo
    even though sources exist at non-default locations in design space. Instances
    at non-default locations must be generatable without tripping the
    is_static_font() assertion in _generate_instance_info.

    https://github.com/googlefonts/ufo2ft/issues/981
    """
    d = designspaceLib.DesignSpaceDocument()
    d.addAxisDescriptor(
        name="Pixel Shape", tag="PXSH", minimum=1, default=1, maximum=100
    )

    # Default source: carries fontinfo, default layer holds the glyph.
    default_font = ufo_module.Font()
    default_font.info.familyName = "Geist Pixel"
    default_font.info.styleName = "Regular"
    default_font.info.unitsPerEm = 1000
    default_font.info.ascender = 800
    default_font.info.descender = -200
    default_font.info.xHeight = 500
    default_font.info.capHeight = 700
    # instance-specific attribute that the multi-master path must NOT inherit
    default_font.info.postscriptFontName = "GeistPixel-Regular"
    default_glyph = default_font.newGlyph("A")
    default_glyph.width = 600
    pen = default_glyph.getPen()
    pen.moveTo((0, 0))
    pen.lineTo((600, 0))
    pen.lineTo((600, 700))
    pen.lineTo((0, 700))
    pen.closePath()

    # Sparse source: shares the same font object but points at a non-default
    # layer. collect_info_masters() will skip it because layerName is not None,
    # so info_mutator ends up with a single master.
    sparse_layer = default_font.newLayer("{100}")
    sparse_glyph = sparse_layer.newGlyph("A")
    sparse_glyph.width = 600
    pen = sparse_glyph.getPen()
    pen.moveTo((50, 50))
    pen.lineTo((550, 50))
    pen.lineTo((550, 650))
    pen.lineTo((50, 650))
    pen.closePath()

    d.addSourceDescriptor(
        name="Geist Pixel Regular",
        familyName="Geist Pixel",
        styleName="Regular",
        location={"Pixel Shape": 1},
        font=default_font,
    )
    d.addSourceDescriptor(
        name="Geist Pixel Regular {100}",
        familyName="Geist Pixel",
        styleName="Regular {100}",
        layerName="{100}",
        location={"Pixel Shape": 100},
        font=default_font,
    )
    d.addInstanceDescriptor(
        familyName="Geist Pixel",
        styleName="Circle",
        location={"Pixel Shape": 20},
    )

    generator = ufo2ft.instantiator.Instantiator.from_designspace(d)

    # info_mutator was built from only the default source (the sparse source was
    # skipped by collect_info_masters), so it reports as a single-master Variator.
    assert generator.info_mutator.is_static_font()

    # The bug: generating an instance at a non-default location currently trips
    # `assert all(v == 0.0 for v in location_normalized.values())`.
    instance_font = generator.generate_instance(d.instances[0])

    # Instance-specific attributes must NOT be inherited from the default source
    # when the instance is at a non-default location. (Only the static-font
    # path that runs at the default location is allowed to inherit them.)
    assert instance_font.info.postscriptFontName is None

    # The interpolating fontinfo (a no-op with a single info master) reduces to
    # the default master's values, which is what we want.
    assert instance_font.info.unitsPerEm == 1000
    assert instance_font.info.ascender == 800
    assert instance_font.info.descender == -200

    # Instance-level overrides still apply.
    assert instance_font.info.familyName == "Geist Pixel"
    assert instance_font.info.styleName == "Circle"


def test_designspace_v5_discrete_axis_raises_error(data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "MutatorSansLite" / "MutatorFamily_v5_discrete_axis.designspace"
    )
    # The error message should advise to use `splitInterpolable()`
    with pytest.raises(
        ufo2ft.instantiator.InstantiatorError, match="splitInterpolable"
    ):
        ufo2ft.instantiator.Instantiator.from_designspace(designspace)


def test_opentype_os2_panose_merging(ufo_module):
    """Test that openTypeOS2Panose values are properly merged across sources.

    Only panose values that are identical across all sources should be copied
    to instances. Different values should be set to 0.
    """
    d = designspaceLib.DesignSpaceDocument()
    d.addAxisDescriptor(
        name="Weight", tag="wght", minimum=300, default=300, maximum=900
    )

    font1 = ufo_module.Font()
    font1.info.openTypeOS2Panose = [2, 11, 5, 2, 4, 5, 4, 2, 2, 4]  # default source

    font2 = ufo_module.Font()
    font2.info.openTypeOS2Panose = [
        2,
        11,
        8,  # different at index 2
        2,
        4,
        5,
        4,
        2,
        2,
        4,
    ]

    font3 = ufo_module.Font()
    font3.info.openTypeOS2Panose = [
        2,
        11,
        5,
        3,  # different at index 3
        4,
        5,
        4,
        2,
        2,
        4,
    ]

    d.addSourceDescriptor(location={"Weight": 300}, font=font1)
    d.addSourceDescriptor(location={"Weight": 600}, font=font2)
    d.addSourceDescriptor(location={"Weight": 900}, font=font3)
    d.addInstanceDescriptor(styleName="Regular", location={"Weight": 400})
    d.addInstanceDescriptor(styleName="Bold", location={"Weight": 700})

    generator = ufo2ft.instantiator.Instantiator.from_designspace(d)

    instance1 = generator.generate_instance(d.instances[0])
    instance2 = generator.generate_instance(d.instances[1])

    # Indices 2 and 3 differ across sources, so they should be 0
    # All other indices are the same across sources
    expected_panose = [2, 11, 0, 0, 4, 5, 4, 2, 2, 4]

    assert instance1.info.openTypeOS2Panose == expected_panose
    assert instance2.info.openTypeOS2Panose == expected_panose


def test_opentype_os2_panose_all_different(ufo_module):
    """Test that when all panose values differ, the attribute is deleted."""
    d = designspaceLib.DesignSpaceDocument()
    d.addAxisDescriptor(
        name="Weight", tag="wght", minimum=300, default=300, maximum=900
    )

    # Create fonts with completely different panose values
    font1 = ufo_module.Font()
    font1.info.openTypeOS2Panose = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    font2 = ufo_module.Font()
    font2.info.openTypeOS2Panose = [10, 9, 8, 7, 6, 5, 4, 3, 2, 1]

    d.addSourceDescriptor(location={"Weight": 300}, font=font1)
    d.addSourceDescriptor(location={"Weight": 900}, font=font2)
    d.addInstanceDescriptor(styleName="Regular", location={"Weight": 600})

    generator = ufo2ft.instantiator.Instantiator.from_designspace(d)
    instance = generator.generate_instance(d.instances[0])

    # All values differ, so openTypeOS2Panose should be unset
    assert instance.info.openTypeOS2Panose is None


def test_opentype_os2_panose_missing_in_some_sources(ufo_module):
    """Test handling when some sources don't have panose values."""
    d = designspaceLib.DesignSpaceDocument()
    d.addAxisDescriptor(
        name="Weight", tag="wght", minimum=300, default=300, maximum=900
    )

    font1 = ufo_module.Font()
    font1.info.openTypeOS2Panose = [2, 11, 5, 2, 4, 5, 4, 2, 2, 4]

    # Second font doesn't have panose, doesn't contribute
    font2 = ufo_module.Font()
    assert font2.info.openTypeOS2Panose is None

    font3 = ufo_module.Font()
    font3.info.openTypeOS2Panose = [2, 11, 5, 2, 4, 5, 4, 2, 2, 4]

    d.addSourceDescriptor(location={"Weight": 300}, font=font1)
    d.addSourceDescriptor(location={"Weight": 600}, font=font2)
    d.addSourceDescriptor(location={"Weight": 900}, font=font3)
    d.addInstanceDescriptor(styleName="Regular", location={"Weight": 400})

    generator = ufo2ft.instantiator.Instantiator.from_designspace(d)
    instance = generator.generate_instance(d.instances[0])

    # The default source's panose should be copied since font2 has None and
    # font3 has same panose as font1
    assert instance.info.openTypeOS2Panose == [2, 11, 5, 2, 4, 5, 4, 2, 2, 4]


def test_opentype_os2_panose_single_source(ufo_module):
    """Test that panose values are preserved as-is when there's only one source.

    Cf. https://github.com/googlefonts/fontc/issues/1609
    """
    d = designspaceLib.DesignSpaceDocument()
    d.addAxisDescriptor(
        name="Weight", tag="wght", minimum=300, default=400, maximum=900
    )

    font1 = ufo_module.Font()
    font1.info.openTypeOS2Panose = [2, 11, 5, 2, 4, 5, 4, 2, 2, 4]

    d.addSourceDescriptor(location={"Weight": 400}, font=font1)
    d.addInstanceDescriptor(styleName="Regular", location={"Weight": 400})

    generator = ufo2ft.instantiator.Instantiator.from_designspace(d)

    instance = generator.generate_instance(d.instances[0])

    # With only one source, panose should be preserved as-is
    assert instance.info.openTypeOS2Panose == [2, 11, 5, 2, 4, 5, 4, 2, 2, 4]


def test_opentype_os2_panose_no_mutation_of_default(ufo_module):
    """Test that processing panose values doesn't mutate the default source."""
    d = designspaceLib.DesignSpaceDocument()
    d.addAxisDescriptor(
        name="Weight", tag="wght", minimum=300, default=300, maximum=900
    )

    font1 = ufo_module.Font()
    original_panose = [2, 11, 5, 2, 4, 5, 4, 2, 2, 4]
    font1.info.openTypeOS2Panose = original_panose.copy()

    font2 = ufo_module.Font()
    font2.info.openTypeOS2Panose = [
        2,
        11,
        8,  # different at index 2
        2,
        4,
        5,
        4,
        2,
        2,
        4,
    ]

    d.addSourceDescriptor(location={"Weight": 300}, font=font1)
    d.addSourceDescriptor(location={"Weight": 900}, font=font2)
    d.addInstanceDescriptor(styleName="Regular", location={"Weight": 600})

    generator = ufo2ft.instantiator.Instantiator.from_designspace(d)
    instance = generator.generate_instance(d.instances[0])

    # The instance should have merged panose with 0 at index 2
    assert instance.info.openTypeOS2Panose == [2, 11, 0, 2, 4, 5, 4, 2, 2, 4]

    # The default source should remain unchanged
    assert font1.info.openTypeOS2Panose == original_panose
    assert d.default.font.info.openTypeOS2Panose == original_panose


@pytest.mark.parametrize(
    "panose_data,expected",
    [
        pytest.param(
            [[2, 11, 5], [2, 11, 8, 2, 4, 5, 4, 2, 2, 4, 99, 88, 77, 66, 55]],
            [2, 11, 0, 0, 0, 0, 0, 0, 0, 0],
            id="zero-padded",  # At least one source has length < 10 (will be padded)
        ),
        pytest.param(
            [
                [2, 11, 5, 2, 4, 5, 4, 2, 2, 4, 99, 88],
                [2, 11, 8, 2, 4, 5, 4, 2, 2, 4, 77, 66, 55],
            ],
            [2, 11, 0, 2, 4, 5, 4, 2, 2, 4],
            id="truncated",  # All sources have length > 10 (will be truncated)
        ),
    ],
)
def test_opentype_os2_panose_malformed_lengths(panose_data, expected, caplog):
    """Test handling of panose values with incorrect lengths (ufoLib2 only)."""
    ufo_module = pytest.importorskip("ufoLib2")

    d = designspaceLib.DesignSpaceDocument()
    d.addAxisDescriptor(
        name="Weight", tag="wght", minimum=300, default=300, maximum=900
    )

    fonts = []
    for i, panose_values in enumerate(panose_data):
        font = ufo_module.Font()
        font.info.openTypeOS2Panose = panose_values
        fonts.append(font)
        d.addSourceDescriptor(location={"Weight": 300 + i * 600}, font=font)

    d.addInstanceDescriptor(styleName="Regular", location={"Weight": 600})

    with caplog.at_level(logging.WARNING):
        generator = ufo2ft.instantiator.Instantiator.from_designspace(d)
        instance = generator.generate_instance(d.instances[0])

    # Should log a warning about invalid length
    assert (
        "openTypeOS2Panose values in designspace sources have invalid length"
        in caplog.text
    )

    # Should produce a valid result with exactly 10 values
    assert instance.info.openTypeOS2Panose == expected
    assert len(instance.info.openTypeOS2Panose) == 10


def test_strict_math_glyph(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "InstantiatorStrictMathGlyph" / "StrictMathGlyph.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )
    fonts = [
        generator.generate_instance(instance) for instance in designspace.instances
    ]
    assert len(fonts) == 1
    glyph = fonts[0]["test"]
    assert len(glyph.contours) == 1
    assert len(glyph.contours[0].points) == 16



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Anchor propagation algorithm ported from:
#   glyphsLib (Apache-2.0): glyphsLib/builder/transformations/propagate_anchors.py

import logging
from collections import deque
from dataclasses import dataclass
from math import atan2, degrees, isinf

from fontTools.misc.transform import Transform

from ufo2ft.constants import (
    _PRELIMINARY_CATEGORIES_KEY,
    GLYPHS_COMPONENT_INFO_KEY,
    OPENTYPE_CATEGORIES_KEY,
)
from ufo2ft.filters import BaseFilter, BaseIFilter
from ufo2ft.util import OpenTypeCategories, _GlyphSet, zip_strict

logger = logging.getLogger(__name__)


@dataclass
class AnchorData:
    name: str
    x: float
    y: float


def _depth_sorted_glyphs(glyph_set):
    """Return glyph names in dependency order (simple glyphs first, then
    composites ordered by increasing component depth).

    Handles missing components and cycles gracefully with warnings.
    """
    queue = deque()
    depths = {}
    waiting_for = {}

    for name in glyph_set:
        glyph = glyph_set[name]
        if not glyph.components:
            depths[name] = 0
        else:
            queue.append(name)

    while queue:
        name = queue.popleft()
        glyph = glyph_set[name]
        comp_names = {c.baseGlyph for c in glyph.components if c.baseGlyph in glyph_set}
        if all(c in depths for c in comp_names):
            depths[name] = max((depths[c] for c in comp_names), default=-1) + 1
            waiting_for.pop(name, None)
        else:
            last_queue_len = waiting_for.get(name)
            waiting_for[name] = len(queue)
            if last_queue_len != len(queue):
                queue.append(name)
            else:
                depths[name] = float("inf")
                waiting_for.pop(name, None)
                logger.warning("glyph '%s' has cyclical components", name)

    for name in glyph_set:
        if name not in depths:
            depths[name] = 0

    return [
        name for _, name in sorted((d, n) for n, d in depths.items() if not isinf(d))
    ]


def get_xy_rotation(transform):
    """Extract scale factors (with sign for flip) from a Transform."""
    a, b = transform[:2]
    angle = atan2(b, a)
    rotated = transform.rotate(-angle)
    xscale, yscale = rotated[0], rotated[3]
    if abs(degrees(angle) - 180) < 0.001:
        xscale = -xscale
        yscale = -yscale
    return xscale, yscale


def rename_anchor_for_scale(name, xscale, yscale):
    """Swap left<->right if xscale<0, bottom<->top if yscale<0,
    exit<->entry if xscale<0."""

    def swap_pair(s, one, two):
        if one in s:
            return s.replace(one, two)
        elif two in s:
            return s.replace(two, one)
        return s

    if xscale < 0.0:
        name = swap_pair(name, "left", "right")
        name = swap_pair(name, "exit", "entry")
    if yscale < 0.0:
        name = swap_pair(name, "bottom", "top")
    return name


def make_liga_anchor_name(name, base_number):
    """Convert 'top' -> 'top_N' for ligature numbering."""
    if "_" in name:
        name, suffix = name.split("_", 1)
        try:
            num = int(suffix)
        except ValueError:
            num = 1
        return f"{name}_{base_number + num}"
    return f"{name}_{base_number + 1}"


def maybe_rename_component_anchor(comp_name, anchors):
    """Handle component attachment anchors like 'top_2'."""
    if "_" not in comp_name:
        return
    sub_name = comp_name[: comp_name.index("_")]
    mark_name = f"_{sub_name}"
    if any(a.name == sub_name for a in anchors) and any(
        a.name == mark_name for a in anchors
    ):
        for a in anchors:
            if a.name == sub_name:
                a.name = comp_name
                break


def origin_adjusted_anchors(anchors):
    """Apply *origin anchor offset and exclude *origin from output."""
    origin_x, origin_y = 0.0, 0.0
    for a in anchors:
        if a.name == "*origin":
            origin_x, origin_y = a.x, a.y
            break
    return [
        AnchorData(name=a.name, x=a.x - origin_x, y=a.y - origin_y)
        for a in anchors
        if a.name != "*origin"
    ]


def _anchors_from_glyph(glyph):
    """Read anchors from a glyph object into AnchorData list."""
    return [AnchorData(name=a.name, x=a.x, y=a.y) for a in glyph.anchors]


def _classify_glyphs(glyph_set, categories):
    """Return (marks, ligatures) sets of glyph names.

    Uses public.openTypeCategories when available, falls back to heuristics.
    """
    marks = set(categories.mark)
    ligatures = set(categories.ligature)

    if categories.is_empty:
        logger.warning(
            "public.openTypeCategories not found or empty; anchor propagation "
            "may differ from fontc/Glyphs.app for mark and ligature glyphs."
        )
        for name in glyph_set:
            glyph = glyph_set[name]
            if any(a.name.startswith("_") for a in glyph.anchors):
                marks.add(name)

    return marks, ligatures


def _finalize_categories(preliminary_categories, done_anchors):
    """Refine preliminary categories after anchor propagation.

    Matches fontc's recompute_gdef_categories: marks are kept as-is,
    ligatures are kept only if they gained attaching anchors, and bases
    are inferred from attaching anchors for all remaining glyphs.
    """
    finalized = {}
    for name, anchors in done_anchors.items():
        has_attaching = any(a.name and not a.name.startswith("_") for a in anchors)
        preliminary = preliminary_categories.get(name)
        if preliminary == "mark":
            finalized[name] = "mark"
        elif preliminary == "ligature" and has_attaching:
            finalized[name] = "ligature"
        elif has_attaching and preliminary != "mark":
            finalized[name] = "base"
    return finalized


def _get_component_anchors(glyph):
    """Read Glyphs component attachment anchors from a glyph's lib.

    Returns a dict mapping component index to the explicit anchor attachment
    name (e.g. {1: "top_1", 2: "top_2"}).
    """
    info = getattr(glyph, "lib", {}).get(GLYPHS_COMPONENT_INFO_KEY, [])
    return {
        entry["index"]: entry["anchor"]
        for entry in info
        if "index" in entry and "anchor" in entry
    }


def anchors_traversing_components(
    glyph_name,
    existing_anchors,
    components,
    is_mark,
    is_ligature,
    done_anchors,
    base_glyph_counts,
    component_anchor_map=None,
):
    """Compute anchors for a glyph, propagating from its components.

    This is a port of the glyphsLib/fontc algorithm adapted for defcon objects.
    """
    if not existing_anchors and not components:
        return []

    if existing_anchors and is_mark:
        return list(origin_adjusted_anchors(existing_anchors))

    has_underscore = any(a.name.startswith("_") for a in existing_anchors)
    number_of_base_glyphs = 0
    all_anchors = {}

    for component_idx, component in enumerate(components):
        comp_name = component.baseGlyph
        if comp_name not in done_anchors:
            logger.warning(
                "Anchors not propagated for inexistent component %s " "in glyph %s",
                comp_name,
                glyph_name,
            )
            continue

        anchors = [
            AnchorData(name=a.name, x=a.x, y=a.y) for a in done_anchors[comp_name]
        ]

        if component_anchor_map and component_idx in component_anchor_map:
            maybe_rename_component_anchor(component_anchor_map[component_idx], anchors)

        component_number_of_base_glyphs = base_glyph_counts.get(comp_name, 0)

        comb_has_underscore = any(
            len(a.name) >= 2 and a.name.startswith("_") for a in anchors
        )
        comb_has_exit = any(a.name.startswith("exit") for a in anchors)
        if not (comb_has_underscore or comb_has_exit):
            all_anchors = {
                n: a for n, a in all_anchors.items() if not n.startswith("exit")
            }

        component_transform = Transform(*component.transformation)
        xscale, yscale = get_xy_rotation(component_transform)

        for anchor in anchors:
            new_has_underscore = anchor.name.startswith("_")
            if (component_idx > 0 or has_underscore) and new_has_underscore:
                continue
            if component_idx > 0 and anchor.name.startswith("entry"):
                continue

            new_anchor_name = rename_anchor_for_scale(anchor.name, xscale, yscale)
            if (
                is_ligature
                and component_number_of_base_glyphs > 0
                and not new_has_underscore
                and not (
                    new_anchor_name.startswith("exit")
                    or new_anchor_name.startswith("entry")
                )
            ):
                new_anchor_name = make_liga_anchor_name(
                    new_anchor_name, number_of_base_glyphs
                )

            pos = component_transform.transformPoint((anchor.x, anchor.y))
            anchor.x = round(pos[0], 6)
            anchor.y = round(pos[1], 6)
            anchor.name = new_anchor_name
            all_anchors[anchor.name] = anchor
            has_underscore |= new_has_underscore

        number_of_base_glyphs += base_glyph_counts.get(comp_name, 0)

    # Apply explicitly defined anchors on this glyph (overrides component ones)
    all_anchors.update((a.name, a) for a in origin_adjusted_anchors(existing_anchors))

    # Count base glyphs from anchors
    has_underscore_anchor = False
    has_mark_anchor = False
    component_count_from_anchors = 0
    for name in all_anchors:
        has_underscore_anchor |= name.startswith("_")
        has_mark_anchor |= bool(name) and name[0].isalpha() and name[0].isascii()
        if (
            not is_ligature
            and number_of_base_glyphs == 0
            and not name.startswith("_")
            and not (name.startswith("exit") or name.startswith("entry"))
            and "_" in name
        ):
            suffix = name[name.index("_") + 1 :]
            maybe_add_one = 1 if name.startswith("caret") else 0
            anchor_index = 0
            try:
                anchor_index = int(suffix) + maybe_add_one
            except ValueError:
                pass
            component_count_from_anchors = max(
                component_count_from_anchors, anchor_index
            )
    if not has_underscore_anchor and number_of_base_glyphs == 0 and has_mark_anchor:
        number_of_base_glyphs += 1
    number_of_base_glyphs = max(number_of_base_glyphs, component_count_from_anchors)

    # Bottom/top cancellation (matches glyphsLib propagate_anchors.py:592-597)
    if any(a.name == "_bottom" for a in existing_anchors):
        all_anchors.pop("top", None)
        all_anchors.pop("_top", None)
    if any(a.name == "_top" for a in existing_anchors):
        all_anchors.pop("bottom", None)
        all_anchors.pop("_bottom", None)

    base_glyph_counts[glyph_name] = number_of_base_glyphs

    return sorted(all_anchors.values(), key=lambda a: a.name)


def _compute_component_closure(glyph_set, root_names):
    """Given a set of included root glyph names, return the full closure
    including all transitive component dependencies."""
    closure = set()
    stack = list(root_names)
    while stack:
        name = stack.pop()
        if name in closure or name not in glyph_set:
            continue
        closure.add(name)
        glyph = glyph_set[name]
        for comp in glyph.components:
            if comp.baseGlyph not in closure and comp.baseGlyph in glyph_set:
                stack.append(comp.baseGlyph)
    return closure


def _combined_glyph_set(glyph_sets):
    """Return a representative glyph set with unioned component dependencies."""
    all_glyph_names = set()
    for glyph_set in glyph_sets:
        all_glyph_names.update(glyph_set.keys())

    combined = {}
    for name in all_glyph_names:
        components = []
        representative = None
        for glyph_set in glyph_sets:
            if name not in glyph_set:
                continue
            glyph = glyph_set[name]
            if representative is None:
                representative = glyph
            components.extend(glyph.components)
        if representative is not None:
            combined[name] = _CombinedGlyph(representative, components)
    return combined


class _CombinedGlyph:
    def __init__(self, glyph, components):
        self._glyph = glyph
        self.components = components

    def __getattr__(self, name):
        return getattr(self._glyph, name)


def _write_anchors_to_glyph(glyph, new_anchors):
    """Write AnchorData list to a glyph, replacing existing anchors.

    Preserves existing anchor identifiers where names match.
    Returns True if the glyph was actually modified.
    """
    old = [(a.name, round(a.x, 6), round(a.y, 6)) for a in glyph.anchors]
    new = [(a.name, round(a.x, 6), round(a.y, 6)) for a in new_anchors]
    if old == new:
        return False

    old_by_name = {}
    for a in glyph.anchors:
        old_by_name.setdefault(a.name, a)

    glyph.clearAnchors()
    for ad in new_anchors:
        anchor_dict = {"name": ad.name, "x": ad.x, "y": ad.y}
        existing = old_by_name.get(ad.name)
        if existing is not None and hasattr(existing, "identifier"):
            ident = getattr(existing, "identifier", None)
            if ident:
                anchor_dict["identifier"] = ident
        try:
            glyph.appendAnchor(anchor_dict)
        except TypeError:
            glyph.appendAnchor(ad.name, (ad.x, ad.y))
    return True


class PropagateAnchorsFilter(BaseFilter):
    # must run before decomposition, while components are still present
    _pre = True

    def __call__(self, font, glyphSet=None):
        fontName = font.__class__.__name__
        logger.info("Running %s on %s", self.name, fontName)

        if glyphSet is None:
            glyphSet = _GlyphSet.from_layer(font)

        categories = OpenTypeCategories.load(font)
        preliminary = None
        if categories.is_empty:
            preliminary = font.lib.get(_PRELIMINARY_CATEGORIES_KEY)
            if preliminary:
                categories = OpenTypeCategories.from_dict(preliminary)
        marks, ligatures = _classify_glyphs(glyphSet, categories)

        sorted_glyphs = _depth_sorted_glyphs(glyphSet)

        done_anchors = {}
        base_glyph_counts = {}

        for name in sorted_glyphs:
            glyph = glyphSet[name]
            existing = _anchors_from_glyph(glyph)
            is_mark = name in marks
            is_ligature = name in ligatures

            result = anchors_traversing_components(
                name,
                existing,
                glyph.components,
                is_mark,
                is_ligature,
                done_anchors,
                base_glyph_counts,
                component_anchor_map=_get_component_anchors(glyph),
            )
            done_anchors[name] = result

        if preliminary:
            finalized = _finalize_categories(preliminary, done_anchors)
            font.lib[OPENTYPE_CATEGORIES_KEY] = finalized
            font.lib.pop(_PRELIMINARY_CATEGORIES_KEY, None)
            logger.info(
                "Finalized preliminary categories: %d base, %d ligature, %d mark",
                sum(1 for v in finalized.values() if v == "base"),
                sum(1 for v in finalized.values() if v == "ligature"),
                sum(1 for v in finalized.values() if v == "mark"),
            )

        # Determine which glyphs to write: included roots + component closure
        include = self.include
        included_roots = {name for name in glyphSet if include(glyphSet[name])}
        write_set = _compute_component_closure(glyphSet, included_roots)

        modified = set()
        for name in sorted_glyphs:
            if name not in write_set:
                continue
            glyph = glyphSet[name]
            if not glyph.components:
                continue
            new_anchors = done_anchors[name]
            if _write_anchors_to_glyph(glyph, new_anchors):
                modified.add(name)

        if modified:
            logger.info("Glyphs with propagated anchors: %i" % len(modified))
        return modified


class PropagateAnchorsIFilter(BaseIFilter):
    _pre = True

    def __call__(self, fonts, glyphSets=None, instantiator=None, **kwargs):
        logger.info("Running interpolatable %s", self.name)

        if glyphSets is None:
            glyphSets = [_GlyphSet.from_layer(font) for font in fonts]

        self.set_context(fonts, glyphSets, instantiator, **kwargs)
        ds_categories = kwargs.get("openTypeCategories")
        preliminary = kwargs.get("preliminaryOpenTypeCategories")
        if ds_categories:
            categories = OpenTypeCategories.from_dict(ds_categories)
        else:
            categories = OpenTypeCategories.load(self.getDefaultFont())
        used_preliminary = categories.is_empty and preliminary
        if used_preliminary:
            categories = OpenTypeCategories.from_dict(preliminary)
        interpolated_layers = self.getInterpolatedLayers()

        # Classify using default font's glyphSet
        default_gs = self.getDefaultGlyphSet()
        marks, ligatures = _classify_glyphs(default_gs, categories)

        # Get union of all glyph names across masters
        all_glyph_names = set()
        for gs in glyphSets:
            all_glyph_names.update(gs.keys())

        combined = _combined_glyph_set(glyphSets)
        sorted_glyphs = _depth_sorted_glyphs(combined)

        # Per-master done_anchors and base_glyph_counts
        per_master_done = [{} for _ in glyphSets]
        per_master_counts = [{} for _ in glyphSets]

        for name in sorted_glyphs:
            is_mark = name in marks
            is_ligature = name in ligatures

            for i, (glyphSet, interpolatedLayer) in enumerate(
                zip_strict(glyphSets, interpolated_layers)
            ):
                if name in glyphSet:
                    glyph = glyphSet[name]
                elif interpolatedLayer is not None and name in interpolatedLayer:
                    glyph = interpolatedLayer[name]
                else:
                    continue

                existing = _anchors_from_glyph(glyph)
                done = per_master_done[i]
                counts = per_master_counts[i]

                result = anchors_traversing_components(
                    name,
                    existing,
                    glyph.components,
                    is_mark,
                    is_ligature,
                    done,
                    counts,
                    component_anchor_map=_get_component_anchors(glyph),
                )
                done[name] = result

        # Finalize preliminary categories after propagation: prune anchorless
        # ligatures and infer bases from attaching anchors.  Write the result
        # to every master UFO's lib so GdefFeatureWriter picks it up.
        if used_preliminary:
            if self.context.instantiator is not None:
                default_idx = self.context.instantiator.default_source_idx
            else:
                default_idx = 0
            default_done = per_master_done[default_idx]
            finalized = _finalize_categories(preliminary, default_done)
            for font in fonts:
                font.lib[OPENTYPE_CATEGORIES_KEY] = finalized
            logger.info(
                "Finalized preliminary categories: %d base, %d ligature, %d mark",
                sum(1 for v in finalized.values() if v == "base"),
                sum(1 for v in finalized.values() if v == "ligature"),
                sum(1 for v in finalized.values() if v == "mark"),
            )

        # Determine which glyphs to write
        include = self.include
        included_roots = set()
        for name in all_glyph_names:
            for gs in glyphSets:
                if name in gs and include(gs[name]):
                    included_roots.add(name)
                    break
        write_set = _compute_component_closure(combined, included_roots)

        modified = set()
        for name in sorted_glyphs:
            if name not in write_set:
                continue
            has_components = any(name in gs and gs[name].components for gs in glyphSets)
            if not has_components:
                continue

            for i, glyphSet in enumerate(glyphSets):
                if name not in glyphSet:
                    continue
                glyph = glyphSet[name]
                new_anchors = per_master_done[i].get(name, [])
                if _write_anchors_to_glyph(glyph, new_anchors):
                    modified.add(name)

        if modified:
            logger.info("Glyphs with propagated anchors: %i" % len(modified))
        return modified



from ufo2ft._compilers.interpolatableOTFCompiler import InterpolatableOTFCompiler
from ufo2ft._compilers.interpolatableTTFCompiler import InterpolatableTTFCompiler
from ufo2ft._compilers.otfCompiler import OTFCompiler
from ufo2ft._compilers.ttfCompiler import TTFCompiler
from ufo2ft._compilers.variableCFF2sCompiler import VariableCFF2sCompiler
from ufo2ft._compilers.variableTTFsCompiler import VariableTTFsCompiler
from ufo2ft.constants import CFFOptimization  # noqa: F401 (fontmake uses it)

__all__ = [
    "compileTTF",
    "compileOTF",
    "compileInterpolatableTTFs",
    "compileVariableTTFs",
    "compileInterpolatableTTFsFromDS",
    "compileInterpolatableOTFsFromDS",
    "compileVariableTTF",
    "compileVariableCFF2",
    "compileVariableCFF2s",
]

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "0.0.0+unknown"


def compileTTF(ufo, **kwargs):
    """Create FontTools TrueType font from a UFO.

    *removeOverlaps* performs a union operation on all the glyphs' contours.

    *flattenComponents* un-nests glyphs so that they have at most one level of
    components.

    *convertCubics* and *cubicConversionError* specify how the conversion from cubic
    to quadratic curves should be handled.

    *layerName* specifies which layer should be compiled. When compiling something
    other than the default layer, feature compilation is skipped.

    *skipExportGlyphs* is a list or set of glyph names to not be exported to the
    final font. If these glyphs are used as components in any other glyph, those
    components get decomposed. If the parameter is not passed in, the UFO's
    "public.skipExportGlyphs" lib key will be consulted. If it doesn't exist,
    all glyphs are exported. UFO groups and kerning will be pruned of skipped
    glyphs.

    *dropImpliedOnCurves* (bool) specifies whether on-curve points that are exactly
    in between two off-curves can be dropped when building glyphs (default: False).

    *allQuadratic* (bool) specifies whether to convert all curves to quadratic - True
    by default, builds traditional glyf v0 table. If False, quadratic curves or cubic
    curves are generated depending on which has fewer points; a glyf v1 is generated.
    """
    return TTFCompiler(**kwargs).compile(ufo)


def compileOTF(ufo, **kwargs):
    """Create FontTools CFF font from a UFO.

    *removeOverlaps* performs a union operation on all the glyphs' contours.

    *optimizeCFF* (int) defines whether the CFF charstrings should be
      specialized and subroutinized. By default both optimization are enabled.
      A value of 0 disables both; 1 only enables the specialization; 2 (default)
      does both specialization and subroutinization.

    *roundTolerance* (float) controls the rounding of point coordinates.
      It is defined as the maximum absolute difference between the original
      float and the rounded integer value.
      By default, all floats are rounded to integer (tolerance 0.5); a value
      of 0 completely disables rounding; values in between only round floats
      which are close to their integral part within the tolerated range.

    *featureWriters* argument is a list of BaseFeatureWriter subclasses or
      pre-initialized instances. Features will be written by each feature
      writer in the given order. If featureWriters is None, the default
      feature writers [KernFeatureWriter, MarkFeatureWriter] are used.

    *filters* argument is a list of BaseFilters subclasses or pre-initialized
      instances. Filters with 'pre' attribute set to True will be pre-filters
      called before the default filters, otherwise they will be post-filters,
      called after the default filters.
      Filters will modify glyphs or the glyph set. The default filters cannot
      be disabled.

    *useProductionNames* renames glyphs in TrueType 'post' or OpenType 'CFF '
      tables based on the 'public.postscriptNames' mapping in the UFO lib,
      if present. Otherwise, uniXXXX names are generated from the glyphs'
      unicode values. The default value (None) will first check if the UFO lib
      has the 'com.github.googlei18n.ufo2ft.useProductionNames' key. If this
      is missing or True (default), the glyphs are renamed. Set to False
      to keep the original names.

    **inplace** (bool) specifies whether the filters should modify the input
      UFO's glyphs, a copy should be made first.

    *layerName* specifies which layer should be compiled. When compiling something
    other than the default layer, feature compilation is skipped.

    *skipExportGlyphs* is a list or set of glyph names to not be exported to the
    final font. If these glyphs are used as components in any other glyph, those
    components get decomposed. If the parameter is not passed in, the UFO's
    "public.skipExportGlyphs" lib key will be consulted. If it doesn't exist,
    all glyphs are exported. UFO groups and kerning will be pruned of skipped
    glyphs.

    *cffVersion* (int) is the CFF format, choose between 1 (default) and 2.

    *subroutinizer* (Optional[str]) is the name of the library to use for
      compressing CFF charstrings, if subroutinization is enabled by optimizeCFF
      parameter. Choose between "cffsubr" or "compreffor".
      By default "cffsubr" is used for both CFF 1 and CFF 2.
      NOTE: cffsubr is required for subroutinizing CFF2 tables, as compreffor
      currently doesn't support it.
    """
    return OTFCompiler(**kwargs).compile(ufo)


def compileInterpolatableTTFs(ufos, **kwargs):
    """Create FontTools TrueType fonts from a list of UFOs with interpolatable
    outlines. Cubic curves are converted compatibly to quadratic curves using
    the Cu2Qu conversion algorithm.

    Return an iterator object that yields a TTFont instance for each UFO.

    *layerNames* refers to the layer names to use glyphs from in the order of
    the UFOs in *ufos*. By default, this is a list of `[None]` times the number
    of UFOs, i.e. using the default layer from all the UFOs.

    When the layerName is not None for a given UFO, the corresponding TTFont object
    will contain only a minimum set of tables ("head", "hmtx", "glyf", "loca", "maxp",
    "post" and "vmtx"), and no OpenType layout tables.

    *skipExportGlyphs* is a list or set of glyph names to not be exported to the
    final font. If these glyphs are used as components in any other glyph, those
    components get decomposed. If the parameter is not passed in, the union of
    all UFO's "public.skipExportGlyphs" lib keys will be used. If they don't
    exist, all glyphs are exported. UFO groups and kerning will be pruned of
    skipped glyphs.
    """
    return InterpolatableTTFCompiler(**kwargs).compile(ufos)


def compileVariableTTFs(designSpaceDoc, **kwargs):
    """Create FontTools TrueType variable fonts for each variable font defined
    in the given DesignSpaceDocument, using their UFO sources
    with interpolatable outlines, using fontTools.varLib.build.

    *optimizeGvar*, if set to False, will not perform IUP optimization on the
      generated 'gvar' table.

    *excludeVariationTables* is a list of sfnt table tags (str) that is passed on
      to fontTools.varLib.build, to skip building some variation tables.

    *variableFontNames* is an optional list of names of variable fonts
      to build. If not provided, all variable fonts listed in the given
      designspace will by built.

    *allQuadratic* (bool) specifies whether to convert all curves to quadratic - True
      by default, builds traditional glyf v0 table. If False, quadratic curves or cubic
      curves are generated depending on which has fewer points; a glyf v1 is generated.

    The rest of the arguments works the same as in the other compile functions.

    Returns a dictionary that maps each variable font filename to a new variable
    TTFont object. If no variable fonts are defined in the Designspace, returns
    an empty dictionary.

    .. versionadded:: 2.28.0
    """
    return VariableTTFsCompiler(**kwargs).compile_variable(designSpaceDoc)


def compileInterpolatableTTFsFromDS(designSpaceDoc, **kwargs):
    """Create FontTools TrueType fonts from the DesignSpaceDocument UFO sources
    with interpolatable outlines. Cubic curves are converted compatibly to
    quadratic curves using the Cu2Qu conversion algorithm.

    If the Designspace contains a "public.skipExportGlyphs" lib key, these
    glyphs will not be exported to the final font. If these glyphs are used as
    components in any other glyph, those components get decomposed. If the lib
    key doesn't exist in the Designspace, all glyphs are exported (keys in
    individual UFOs are ignored). UFO groups and kerning will be pruned of
    skipped glyphs.

    The DesignSpaceDocument should contain SourceDescriptor objects with 'font'
    attribute set to an already loaded defcon.Font object (or compatible UFO
    Font class). If 'font' attribute is unset or None, an AttributeError exception
    is thrown.

    Return a copy of the DesignSpaceDocument object (or the same one if
    inplace=True) with the source's 'font' attribute set to the corresponding
    TTFont instance.

    For sources that have the 'layerName' attribute defined, the corresponding TTFont
    object will contain only a minimum set of tables ("head", "hmtx", "glyf", "loca",
    "maxp", "post" and "vmtx"), and no OpenType layout tables.
    """
    return InterpolatableTTFCompiler(**kwargs).compile_designspace(designSpaceDoc)


def compileInterpolatableOTFsFromDS(designSpaceDoc, **kwargs):
    """Create FontTools CFF fonts from the DesignSpaceDocument UFO sources
    with interpolatable outlines.

    Interpolatable means without subroutinization and specializer optimizations
    and no removal of overlaps.

    If the Designspace contains a "public.skipExportGlyphs" lib key, these
    glyphs will not be exported to the final font. If these glyphs are used as
    components in any other glyph, those components get decomposed. If the lib
    key doesn't exist in the Designspace, all glyphs are exported (keys in
    individual UFOs are ignored). UFO groups and kerning will be pruned of
    skipped glyphs.

    The DesignSpaceDocument should contain SourceDescriptor objects with 'font'
    attribute set to an already loaded defcon.Font object (or compatible UFO
    Font class). If 'font' attribute is unset or None, an AttributeError exception
    is thrown.

    Return a copy of the DesignSpaceDocument object (or the same one if
    inplace=True) with the source's 'font' attribute set to the corresponding
    TTFont instance.

    For sources that have the 'layerName' attribute defined, the corresponding TTFont
    object will contain only a minimum set of tables ("head", "hmtx", "CFF ", "maxp",
    "vmtx" and "VORG"), and no OpenType layout tables.
    """
    return InterpolatableOTFCompiler(**kwargs).compile_designspace(designSpaceDoc)


def compileVariableTTF(designSpaceDoc, **kwargs):
    """Create FontTools TrueType variable font from the DesignSpaceDocument UFO sources
    with interpolatable outlines, using fontTools.varLib.build.

    *optimizeGvar*, if set to False, will not perform IUP optimization on the
      generated 'gvar' table.

    *excludeVariationTables* is a list of sfnt table tags (str) that is passed on
      to fontTools.varLib.build, to skip building some variation tables.

    The rest of the arguments works the same as in the other compile functions.

    Returns a new variable TTFont object.
    """
    fonts = VariableTTFsCompiler(**kwargs).compile_variable(designSpaceDoc)
    if len(fonts) != 1:
        raise ValueError(
            "Tried to build a DesignSpace version 5 with multiple variable "
            "fonts using the old ufo2ft API `compileVariableTTF`. "
            "Use the new API instead `compileVariableTTFs`"
        )
    return next(iter(fonts.values()))


def compileVariableCFF2(designSpaceDoc, **kwargs):
    fonts = VariableCFF2sCompiler(**kwargs).compile_variable(designSpaceDoc)
    if len(fonts) != 1:
        raise ValueError(
            "Tried to build a DesignSpace version 5 with multiple variable "
            "fonts using the old ufo2ft API `compileVariableCFF2`. "
            "Use the new API instead `compileVariableCFF2s`"
        )
    return next(iter(fonts.values()))


def compileVariableCFF2s(designSpaceDoc, **kwargs):
    """Create FontTools TrueType variable fonts for each variable font defined
    in the given DesignSpaceDocument, using their UFO sources
    with interpolatable outlines, using fontTools.varLib.build.

    *optimizeGvar*, if set to False, will not perform IUP optimization on the
      generated 'gvar' table.

    *excludeVariationTables* is a list of sfnt table tags (str) that is passed on
      to fontTools.varLib.build, to skip building some variation tables.

    *variableFontNames* is an optional list of names of variable fonts
      to build. If not provided, all variable fonts listed in the given
      designspace will by built.

    *allQuadratic* (bool) specifies whether to convert all curves to quadratic - True
      by default, builds traditional glyf v0 table. If False, quadratic curves or cubic
      curves are generated depending on which has fewer points; a glyf v1 is generated.

    The rest of the arguments works the same as in the other compile functions.

    Returns a dictionary that maps each variable font filename to a new variable
    TTFont object. If no variable fonts are defined in the Designspace, returns
    an empty dictionary.

    .. versionadded:: 2.28.0
    """
    return VariableCFF2sCompiler(**kwargs).compile_variable(designSpaceDoc)



from dataclasses import dataclass
from typing import Optional, Type

from ufo2ft.outlineCompiler import OutlineTTFCompiler
from ufo2ft.preProcessor import TTFPreProcessor
from ufo2ft.util import prune_unknown_kwargs

from .baseCompiler import BaseCompiler


@dataclass
class TTFCompiler(BaseCompiler):
    preProcessorClass: Type = TTFPreProcessor
    outlineCompilerClass: Type = OutlineTTFCompiler
    convertCubics: bool = True
    cubicConversionError: Optional[float] = None
    reverseDirection: bool = True
    rememberCurveType: bool = True
    flattenComponents: bool = False
    autoUseMyMetrics: bool = True
    dropImpliedOnCurves: bool = False
    allQuadratic: bool = True

    def compileOutlines(self, ufo, glyphSet):
        kwargs = prune_unknown_kwargs(self.__dict__, self.outlineCompilerClass)
        kwargs["glyphDataFormat"] = 0 if self.allQuadratic else 1
        outlineCompiler = self.outlineCompilerClass(ufo, glyphSet=glyphSet, **kwargs)
        return outlineCompiler.compile()
